#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Top-level module for carat"""

import warnings
import re
from .version import version as __version__
from .version import show_versions

# And all the carat sub-modules
#from ._cache import cache
from . import util
from . import audio
from . import features
from . import annotations
from . import clustering
from . import display

# Exporting exception classes at the top level
#from .util.exceptions import *  # pylint: disable=wildcard-import

# Exporting all core functions is okay here: suppress the import warning
#from .core import *  # pylint: disable=wildcard-import

warnings.filterwarnings('always',
                        category=DeprecationWarning,
                        module='^{0}'.format(re.escape(__name__)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Version info"""

import sys
import importlib

short_version = '0.1'
version = '0.1.0'


def __get_mod_version(modname):

    try:
        if modname in sys.modules:
            mod = sys.modules[modname]
        else:
            mod = importlib.import_module(modname)
        try:
            return mod.__version__
        except AttributeError:
            return 'installed, no version number available'

    except ImportError:
        return None


def show_versions():
    '''Return the version information for all carat dependencies.'''

    core_deps = ['audioread',
                 'numpy',
                 'scipy',
                 'sklearn']

    extra_deps = ['numpydoc',
                  'sphinx',
                  'matplotlib']

    print('INSTALLED VERSIONS')
    print('------------------')
    print('python: {}\n'.format(sys.version))
    print('carat: {}\n'.format(version))
    for dep in core_deps:
        print('{}: {}'.format(dep, __get_mod_version(dep)))
    print('')
    for dep in extra_deps:
        print('{}: {}'.format(dep, __get_mod_version(dep)))
    pass



# encoding: utf-8
# pylint: disable=C0103
"""Utility functions to deal with audio."""

import librosa

__all__ = ['load']

# simply use librosa.load (this may change in the future)
load = librosa.load




# encoding: utf-8
# pylint: disable=C0103
# pylint: disable=too-many-arguments
"""Utility functions to deal with audio."""

from sklearn.cluster import KMeans

__all__ = ['rhythmic_patterns']

def rhythmic_patterns(data, n_clusters=4, method='kmeans'):
    """Clustering of rhythmic patterns from feature map.

    Based on the feature map clustering analysis introduced in [1].

    [1] Rocamora, Jure, Biscainho
           "Tools for detection and classification of piano drum patterns from candombe recordings."
           9th Conference on Interdisciplinary Musicology (CIM),
           Berlin, Germany. 2014.

    **Args**:
        - data (numpy array): feature map
        - n_clusters (int): number of clusters
        - method (str): clustering method
    **Returns**:
        - clusters(numpy array):

    **Raises**:
        -
    """

    if method == 'kmeans':
        # initialize k-means algorithm
        kmeans = KMeans(n_clusters=n_clusters, init='k-means++', n_init=10)

        # cluster data using k-means
        kmeans.fit(data)

        # predict cluster for each data point
        clusters = kmeans.predict(data)

    else:
        raise AttributeError("Clustering method not implemented.")

    return clusters



from setuptools import setup, find_packages
from importlib.machinery import SourceFileLoader
import sys


version = SourceFileLoader('carat.version',
                               'carat/version.py').load_module()

with open('README.md', 'r') as fdesc:
    long_description = fdesc.read()

setup(
    name='carat',
    version=version.version,
    description='Computer-Aided Rhythm Analysis Toolbox',
    author='Martín Rocamora',
    author_email='rocamora@fing.edu.uy',
    url='http://github.com/mrocamora/carat',
    download_url='http://github.com/mrocamora/carat/releases',
    packages=find_packages(),
    package_data={'': ['example_data/*']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='audio music sound',
    license='MIT',
    install_requires=[
        'audioread >= 2.0.0',
        'numpy >= 1.8.0',
        'scipy >= 1.0.0',
        'scikit-learn >= 0.14.0',
        'soundfile >= 0.9.0',
    ],
    extras_require={
        'docs': ['numpydoc',
                 'matplotlib >= 2.0.0'],
        'tests': ['matplotlib >= 2.1'],
        'display': ['matplotlib >= 1.5'],
    }
)



# encoding: utf-8
# pylint: disable=C0103
# pylint: disable=too-many-arguments
"""Functions to compute accentuation features and feature maps."""

import numpy as np
import scipy as sp
from . import util

__all__ = ['accentuation_feature']

def accentuation_feature(signal, fs, sum_flag=True, log_flag=False, mel_flag=True,
                         alpha=1000, maxfilt_flag=False, maxbins=3, **kwargs):
    """Compute accentuation feature from audio signal.


    Based on the log-power Mel spectrogram [1].

    [1] Böck, Sebastian, and Gerhard Widmer.
           "Maximum filter vibrato suppression for onset detection."
           16th International Conference on Digital Audio Effects,
           Maynooth, Ireland. 2013.

        This performs the following calculations to the input signal:

        input->STFT->(Mel scale)->(Log)->(Max filtering)->Diff->HWR->(Sum)

        Parenthesis denote optional steps.

    **Args**:
        - input: signal
        - fs: sampling rate
        - sum_flag (bool): true if the features are to be summed for each frame.
        - log_flag (bool): true if the features energy are to be converted to dB.
        - mel_flag (bool): true if the features are to be mapped in the Mel scale.
        - alpha (int): compression parameter for dB conversion - log10(alpha*abs(S)+1).
        - maxfilt_flag (bool): true if a maximum filtering is applied to the feature.
        - maxbins (int): number of frequency bins for maximum filter size
        - **kw: these keyword arguments are passed down to each of the functions used
    **Returns**:
        - feature (numpy array): feature values
        - time (numpy array): time values

    **Raises**:
        -
    """

    # STFT
    val_kw, remaining_kw = util.getValidKeywords(kwargs, spectrogram)
    feature, time, frequency = spectrogram(signal, fs, **val_kw)
    # mel scale mapping (and absolute value)
    if mel_flag:
        val_kw, remaining_kw = util.getValidKeywords(remaining_kw, melSpectrogram)
        feature, time, frequency = melSpectrogram(feature, time, frequency, **val_kw)
    else:
        # take the absolute value
        feature = sp.absolute(feature)
    # log magnitude (with compression parameter)
    if log_flag:
        feature = 20*sp.log10(alpha * feature + 1)
    # maximum filter (and difference)
    if maxfilt_flag:
        # maximum filter
        max_spec = sp.ndimage.filters.maximum_filter(feature, size=(maxbins, 1))
        # init the diff array
        diff = sp.zeros(feature.shape)
        # calculate difference between log spec and max filtered version
        diff[:, 1:] = (feature[:, 1:] - max_spec[:, : -1])
        # save feature data
        feature = diff
    else:
        # conventional difference (first order)
        feature = calculateDelta(feature, delta_filter_length=1)
    # half-wave rectification
    feature = halfWaveRectification(feature)
    # sum features
    if sum_flag:
        feature = sumFeatures(feature)

    # return
    return feature, time, frequency


def feature_map(feature, time, beats, downbeats, n_beats=4, n_tatums=4,
                norm_flag=True, pnorm=8, window=0.1):
    """Compute feature map from accentuation feature signal.


    Based on the feature map introduced in [1].

    [1] Rocamora, Jure, Biscainho
           "Tools for detection and classification of piano drum patterns from candombe recordings."
           9th Conference on Interdisciplinary Musicology (CIM),
           Berlin, Germany. 2014.


    The accentuation feature is organized into a feature map. First, the feature signal is
    time-quantized to the rhythm metric structure by considering a grid of tatum pulses equally
    distributed within the annotated beats. The corresponding feature value is taken as the maximum
    within window centered at the frame closest to each tatum instant. This yields feature vectors
    whose coordinates correspond to the tatum pulses of the rhythm cycle (or bar). Finally, a
    feature map of the cycle-length rhythmic patterns of the audio file is obtained by building a
    matrix whose columns are consecutive feature vectors.

    **Args**:
        - feature (numpy array): feature signal
        - **kw: these keyword arguments are passed down to each of the functions used
    **Returns**:
        - :

    **Raises**:
        -
    """

    normalized_feature = np.copy(feature)

    if norm_flag:
        # normalize feature values with a p-norm applied within a local window
        normalized_feature = feature_normalization(feature, time, beats,
                                                   n_tatums=n_tatums, pnorm=pnorm)

    # generate tatum grid to time quantize feature values
    tatums = generate_tatum_grid(beats, downbeats, n_tatums)

    # time quantize the feature signal to the tatum grid
    quantized_feature = feature_time_quantize(normalized_feature, time, tatums, window=window)

    # reshape into a matrix whose columns are bars and its elements are tatums
    features_map = np.reshape(quantized_feature, (-1, n_beats*n_tatums))

    return features_map, quantized_feature, tatums, normalized_feature


def feature_normalization(feature, time, beats, n_tatums=4, pnorm=8):
    """Local amplitude normalization of the feature signal.

    Based on the feature map introduced in [1] and detailed in [2].

    [1] Rocamora, Jure, Biscainho
           "Tools for detection and classification of piano drum patterns from candombe recordings."
           9th Conference on Interdisciplinary Musicology (CIM),
           Berlin, Germany. 2014.

    [2] Rocamora, Cancela, Biscainho
           "Information theory concepts applied to the analysis of rhythm in recorded music with
           recurrent rhythmic patterns."
           Journal of the AES, 67(4), 2019.

    A local amplitude normalization is carried out to preserve intensity variations of the
    rhythmic patterns while discarding long-term fluctuations in dynamics. A p-norm within
    a local window is applied. The window width is proportional to the beat period.

    **Args**:
        - feature (numpy array): feature signal values
        - time (numpy array): time instants of the feature values
        - beats (numpy array): time instants of the tactus beats
        - n_tatums (int): number of tatums per tactus beat
        - pnorm (int): p-norm order for normalization
    **Returns**:
        - :

    **Raises**:
        - norm_feature (numpy array): normalized feature signal values
    """

    # estimate tatum period from annotations for normalization
    # time sample interval
    samps_time = time[1] - time[0]
    # compute tatum period in seconds from median of the tactus intervals
    beats_periods = beats[1:] - beats[0:-1]
    # tatum period in seconds
    tatum_period_secs = np.median(beats_periods) / n_tatums
    # period in samples
    tatum_period_samps = int(round(tatum_period_secs / samps_time)) - 1
    # normalize feature
    norm_feature = normalize_features(feature, tatum_period_samps * pnorm, p=pnorm)

    return norm_feature


def generate_tatum_grid(beats, downbeats, n_tatums):
    """Generate tatum temporal grid from time instants of the tactus beats.

    A grid of tatum pulses is generated equally distributed within the given tactus beats.
    The grid is used to time quantize the feature signal to the rhythmic metric structure.

    Parameters
    ----------
    labels_time (np.ndarray) : time instants of the tactus beats
    labels (list)            : labels at the tactus beats (e.g. 1.1, 1.2, etc)

    Returns
    -------
    tatum_time (np.ndarray)  : time instants of the tatum beats
    """

    # first and last downbeat
    first_downbeat = downbeats[0]
    last_downbeat = downbeats[-1]

    # index of first and last downbeat into the beats array
    # NOTE: we assume coincident beats and downbeats (because of our data)
    indx_first = int(np.where(beats == first_downbeat)[0])
    indx_last = int(np.where(beats == last_downbeat)[0])

    # number of tactus beats
    num_beats = indx_last - indx_first # the beat of the last downbeat is not counted

    # time instants of the tatums
    tatums = np.zeros(num_beats * n_tatums)

    # compute tatum time locations from the tactus beats
    for ind in range(indx_first, indx_last):
        # tatum period estimated from tactus beats
        tatum_period = (beats[ind+1] - beats[ind]) / n_tatums
        # a whole bar of tatum beats
        tatum_bar = np.array(range(n_tatums) * tatum_period + beats[ind])
        # save bar of tatum beats
        tatums[(ind*n_tatums):((ind+1)*n_tatums)] = tatum_bar

    return tatums


def feature_time_quantize(feature, time, tatums, window=0.1):
    """Time quantization of the feature signal to a tatum grid.

    The feature signal is time-quantized to the rhythm metric structure by considering a grid of
    tatum pulses equally distributed within the tactus beats. The feature value assigned to each
    tatum time instant is obtained as the maximum value of the feature signal within a certain
    window centered at the tatum time instant. Default value for the total window lenght is 100 ms.

    **Args**:
        - feature (numpy array): feature signal values
        - time (numpy array): time instants of the feature values
        - tatums (numpy array): time instants of the tatum grid
    **Returns**:
        - :

    **Raises**:
        - quantized_feature (numpy array): time quantized feature signal values
    """
    # number of tatum instants
    num_tatums = tatums.size

    # time quantized feature values at tatum instants
    quantized_feature = np.zeros(tatums.shape)

    # compute hop size to set time-quantization window
    hop_size = time[1] - time[0]
    # number of frames within time quantization window
    win_num_frames = int(window / hop_size)
    # half window in frames
    hwf = int(np.floor(win_num_frames / 2))
    # check if it is even to center window
    if hwf % 2 != 0:
        hwf -= 1

    # compute feature value considering a certain neighbourhood
    for ind in range(num_tatums):
        # closest feature frame to tatum instant
        ind_tatum = util.find_nearest(time, tatums[ind])
        # get maximum feature value within a neighbourhood
        if ind_tatum == 0:
            quantized_feature[ind] = np.max(feature[ind_tatum:ind_tatum+hwf+1])
        else:
            quantized_feature[ind] = np.max(feature[ind_tatum-hwf:ind_tatum+hwf+1])

    return quantized_feature

def spectrogram(signal, fs, window_length=20e-3, hop=10e-3,
                windowing_function=sp.hanning, dft_length=None, zp_flag=False):
    """ Calculates the Short-Time Fourier Transform a signal.

    Given an input signal, it calculates the DFT of frames of the signal and stores them
    in bi-dimensional Scipy array.

    **Args**:
        - window_len (float):length of the window in seconds (must be positive).
        - window (callable): a callable object that receives the window length in samples
          and returns a numpy array containing the windowing function samples.
        - hop (float): frame hop between adjacent frames in seconds.
        - zp_flag (bool): a flag indicating if the *Zero-Phase Windowing* should be
          performed.

    **Returns**:
        - spec (numpy array): spectrogram data
        - time (numpy array): time in seconds of each frame
        - frequnecy (numpy array): frequency grid

    """
    # Converting window_length and hop from seconds to number of samples:
    win_samps = int(round(window_length * fs))
    hop_samps = max(int(round(hop * fs)), 1)
    spec, time, frequency = util.STFT(signal, win_samps, hop_samps,
                                      windowing_function, dft_length, zp_flag)
    # convert indices to seconds and Hz
    time /= fs
    frequency *= fs
    # return
    return spec, time, frequency


def melSpectrogram(in_spec, in_time, in_freq, nfilts=40, minfreq=20, maxfreq=None):
    """This function converts a Spectrogram with linearly spaced frequency components
    to the Mel scale.

        Given an input signal, it calculates the DFT of frames of the signal and stores
        them in bi-dimensional Scipy array.


    **Args**:
        - window_len (float): length of the window in seconds (must be positive).
        - window (callable): a callable object that receives the window length in samples
          and returns a numpy array containing the windowing function samples.
    **Returns**:
        - spec (numpy array): mel-spectrogram data
        - time (numpy array): time in seconds of each frame
        - frequnecy (numpy array): frequency grid

    """
    if maxfreq is None:
        maxfreq = max(in_freq)
    (wts, frequency) = util.fft2mel(in_freq, nfilts, minfreq, maxfreq)
    spec = sp.dot(wts, sp.sqrt(sp.absolute(in_spec)**2))
    time = in_time

    # return
    return spec, time, frequency


def halfWaveRectification(in_signal):
    """ Half-wave rectifies features.

        All feature values below zero are assigned to zero.

    **Args**:
        - input: feature object
        - delta_filter_length (int): length of the filter used to calculate the Delta
          coefficients. Must be an odd number.
    **Returns**:
        - output: numpy array
        A :class:`Feature` object containing the result of the calculation.

    **Raises**:
        - ValueError when the input features are complex.
    """
    out_signal = np.copy(in_signal)
    if out_signal.dtype != complex:
        out_signal[out_signal < 0] = 0.0
    else:
        raise ValueError('Cannot half-wave rectify a complex signal.')
    return out_signal


def calculateDelta(in_signal, delta_filter_length=3):
    """ This function calculates the delta coefficients of a given feature.

    **Args**:
        - input: input feature signal
        - delta_filter_length (int): length of the filter used to calculate the Delta
          coefficients. Must be an odd number.
    **Returns**:
        - output: output feature signal
    """
    out_signal = np.copy(in_signal)
    out_signal = util.deltas(out_signal, delta_filter_length)
    return out_signal


def sumFeatures(in_signal):
    """ This function sums all features along frames.

    **Args**:
        - input: input feature signal
    **Returns**:
        - output: output feature signal
    """
    out_signal = np.copy(in_signal)
    out_signal = sp.sum(out_signal, axis=0)
    return out_signal


def def_norm_feat_gen(data, max_period, p):
    """ Normalization of the feature signal using p-norm.
    """
    if not max_period  % 2:
        max_period += 1
    ext_len = int((max_period - 1) / 2)
    ext_data = data[1:ext_len + 1][::-1]
    ext_data = sp.append(ext_data, data)
    ext_data = sp.append(ext_data, data[-2:-ext_len - 2:-1])

    def aux(i, win_size):
        fac = int(win_size % 2)
        h_len = int(win_size / 2)
        aux = sp.linalg.norm(ext_data[i - h_len + ext_len : i + ext_len + h_len + fac], ord=p)
        return ext_data[i + ext_len] / max(aux, 1e-20)

    return aux


def normalize_features(data, win_len, p):
    """ Normalization of the feature signal using p-norm.
    """
    aux = def_norm_feat_gen(data, win_len, p)
    out = data.copy()
    for i in range(data.size):
        out[i] = aux(i, win_len)
    return out


#def accentuation_feature(y, sr=22050, hop_length=512, n_fft=2048,
#                         n_mels=128, freq_sb=None, **kwargs):
#    """Compute accentuation feature from audio signal.
#
#
#    Based on the log-power Mel spectrogram [1].
#
#    [1] Böck, Sebastian, and Gerhard Widmer.
#           "Maximum filter vibrato suppression for onset detection."
#           16th International Conference on Digital Audio Effects,
#           Maynooth, Ireland. 2013.
#
#    In current implementation win_length = n_fft (because of librosa)
#    The log-power Mel spectrogram is computed for the full spectrum,
#    i.e. up to sr/2 but fmin and fmax could be used to limit the representation.
#
#    A frequency band to focus on.
#
#    Parameters
#    ----------
#
#    Returns
#    -------
#
#    Examples
#    --------
#
#    Notes
#    -----
#    """
#
#    if freq_sb is None:
#        # compute feature values for the full spectrum
#        feature_values = librosa.onset.onset_strength(y=y, sr=sr, n_fft=n_fft, n_mels=n_mels,
#                                                      hop_length=hop_length, **kwargs)
#    else:
#        # check if two frequency values are provided
#        if isinstance(freq_sb, np.ndarray) and freq_sb.shape[0] == 2:
#
#            # compute the frequency of the mel channels
#            n_freqs = librosa.core.time_frequency.mel_frequencies(n_mels=n_mels, fmin=0.0,
#                                                                  fmax=sr/2, htk=False)
#            # find indexes of sub-band channels
#            channels = [util.find_nearest(n_freqs, freq) for freq in freq_sb]
#            # compute feature values for the full spectrum and aggregate across a sub-band
#            feature_values = librosa.onset.onset_strength_multi(y=y, sr=sr,
#                                                                n_fft=n_fft,
#                                                                n_mels=n_mels,
#                                                                hop_length=hop_length,
#                                                                channels=channels,
#                                                                **kwargs)
#            # save a single sub-band
#            feature_values = feature_values[0]
#
#    # compute time instants of the feature values
#    times = librosa.frames_to_time(np.arange(feature_values.shape[0]), sr=sr,
#                                   n_fft=n_fft, hop_length=hop_length)
#    # remove offset of n_fft/2 and hop_length because of the time lag for computing differences
#    times = times - (n_fft/2/sr) - (hop_length/sr)
#    # times = (np.arange(feature_values.shape[0]) * hop_length + win_length/2) / sr
#
#    return feature_values, times



# encoding: utf-8
# pylint: disable=C0103
"""Utility functions to deal with annotations."""

import csv
import numpy as np

__all__ = ['load_beats', 'load_downbeats']

def load_beats(labels_file, delimiter=',', times_col=0, labels_col=1):
    """Load annotated beats from text (csv) file.

    Parameters
    ----------
    labels_file (str) : name (including path) of the input file
    times_col (int) : column index of the time data
    labels_col (int) : column index of the label data

    Returns
    -------
    beat_times  (np.ndarray) : time instants of the beats
    beat_labels (list)       : labels at the beats (e.g. 1.1, 1.2, etc)

    Examples
    --------
    >>> # load an example file from the candombe dataset
    >>> # (http://www.eumus.edu.uy/candombe/datasets/ISMIR2015/)
    >>> beats, beat_labs = annotations.load_beats('candombe_beats.csv')
    >>> beats[0]
    0.548571428
    >>> beat_labs[0]
    '1.1'

    >>> # load an example file from the samba dataset
    >>> beats, beat_labs = annotations.load_beats('./material/data/samba/[0216]
                                                  S2-TB2-03-SE.beats.txt', delimiter=' ')
    >>> beats
    array([ 2.088,  2.559,  3.012,  3.48 ,  3.933,  4.41 ,  4.867,  5.32 ,
            5.771,  6.229,  6.69 ,  7.167,  7.633,  8.092,  8.545,  9.01 ,
            9.48 ,  9.943, 10.404, 10.865, 11.322, 11.79 , 12.251, 12.714,
           13.167, 13.624, 14.094, 14.559, 15.014, 15.473, 15.931, 16.4  ,
           16.865, 17.331, 17.788, 18.249, 18.706, 19.167, 19.643, 20.096,
           20.557, 21.018, 21.494, 21.945, 22.408, 22.869, 23.31 , 23.773,
           24.235, 24.692, 25.151, 25.608, 26.063, 26.52 ])
    >>> beat_labs
    ['1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2']


    Notes
   -----
    It is assumed that the beat annotations are provided as a text file (csv).
    Apart from the time data (mandatory) a label can be given for each beat (optional).
    The time data is assumed to be given in seconds.
    The labels may indicate the beat number within the rhythm cycle (e.g. 1.1, 1.2, or 1, 2).
    """

    # read beat time instants
    beat_times = np.genfromtxt(labels_file, delimiter=delimiter, usecols=(times_col))

    # read beat labels
    with open(labels_file, 'r') as fi:
        reader = csv.reader(fi, delimiter=delimiter)
        # number of columns
        ncol = len(next(reader))
        # check if there are no labels
        if ncol == 1:
            beat_labels = []
        else:
            fi.seek(0)
            beat_labels = [row[labels_col] for row in reader]

    return beat_times, beat_labels


def load_downbeats(labels_file, delimiter=',', times_col=0, labels_col=1, downbeat_label='.1'):
    """Load annotated downbeats from text (csv) file.


    Parameters
    ----------
    labels_file (str) : name (including path) of the input file
    times_col (int) : column index of the time data
    labels_col (int) : column index of the label data
    downbeat_label (str) : string to look for in the label data to select downbeats

    Returns
    -------
    downbeat_times  (np.ndarray) : time instants of the downbeats
    downbeat_labels (list)       : labels at the downbeats

    Examples
    --------
    >>> # Load an example labels file
    >>> y, sr = librosa.load(librosa.util.example_audio_file())
    >>> beats, beat_labs = annotations.load_beats('candombe_beats.csv')
    >>> librosa.get_duration(y=y, sr=sr)
    61.45886621315193


    >>> # load an example file from the samba dataset
    >>> beats, beat_labs = annotations.load_beats('./material/data/samba/[0216]
                                                  S2-TB2-03-SE.beats.txt', delimiter=' ')
    >>> beats
    array([ 2.088,  2.559,  3.012,  3.48 ,  3.933,  4.41 ,  4.867,  5.32 ,
            5.771,  6.229,  6.69 ,  7.167,  7.633,  8.092,  8.545,  9.01 ,
            9.48 ,  9.943, 10.404, 10.865, 11.322, 11.79 , 12.251, 12.714,
           13.167, 13.624, 14.094, 14.559, 15.014, 15.473, 15.931, 16.4  ,
           16.865, 17.331, 17.788, 18.249, 18.706, 19.167, 19.643, 20.096,
           20.557, 21.018, 21.494, 21.945, 22.408, 22.869, 23.31 , 23.773,
           24.235, 24.692, 25.151, 25.608, 26.063, 26.52 ])
    >>> beat_labs
    ['1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2',
     '1', '2', '1', '2', '1', '2', '1', '2', '1', '2', '1', '2']

    >>> downbeats, downbeat_labs = annotations.load_downbeats('./material/data/samba/[0216]
                                                              S2-TB2-03-SE.beats.txt', delimiter='
                                                              ', downbeat_label='1')
    >>> downbeats
    array([ 2.088,  3.012,  3.933,  4.867,  5.771,  6.69 ,  7.633,  8.545,
            9.48 , 10.404, 11.322, 12.251, 13.167, 14.094, 15.014, 15.931,
           16.865, 17.788, 18.706, 19.643, 20.557, 21.494, 22.408, 23.31 ,
           24.235, 25.151, 26.063])
    >>> downbeat_labs
    ['1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1',
     '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1', '1']


    Notes
    -----
    It is assumed that the annotations are provided as a text file (csv).
    Apart from the time data (mandatory) a label can be given for each downbeat (optional).
    The time data is assumed to be given in seconds.

    If a single file contains both beats and downbeats then the downbeat_label is used to select
    downbeats. The downbeats are those beats whose label has the given downbeat_label string. For
    instance the beat labels can be numbers, e.g. '1', '2'. Then, the downbeat_label is just '1'.
    This is the case for the BRID samba dataset. In the case of the candombe dataset, the beat
    labels indicate bar number and beat number. For instance, '1.1', '1.2', '1.3' and '1.4' are the
    four beats of the first bar. Hence, the string needed to indetify the downbeats is '.1'.
    """

    # read file as beats
    beat_times, beat_labs = load_beats(labels_file, delimiter=delimiter, times_col=times_col,
                                       labels_col=labels_col)

    # if there are no labels in the file or downbeat_label is None, then all entries are downbeats
    if not beat_labs or downbeat_label is None:
        downbeat_times, downbeat_labs = beat_times, beat_labs
    else:
        # get downbeat instants and labels by finding the string downbeat_label in the beat labels
        ind_downbeats = [ind_beat for ind_beat in range(len(beat_labs)) if downbeat_label in
                         beat_labs[ind_beat]]
        downbeat_times = beat_times[ind_downbeats]
        downbeat_labs = [beat_labs[ind_downbeat] for ind_downbeat in ind_downbeats]

    return downbeat_times, downbeat_labs



# encoding: utf-8
# pylint: disable=C0103
# pylint: disable=too-many-arguments
"""Utility functions."""

import warnings
import numpy as np
import scipy as sp
import scipy.signal
import scipy.fftpack as fft
#import scipy.signal
#import exceptions


__all__ = ['find_nearest', 'STFT', 'hz2mel', 'mel2hz', 'deltas']


def find_nearest(array, value):
    """Find index of the nearest value of an array to a given value

    Parameters
    ----------
    array (numpy.ndarray)  : array
    value (float)          : value

    Returns
    -------
    idx (int)              : index of nearest value in the array
    """

    idx = (np.abs(array-value)).argmin()

    return idx


def STFT(x, window_length, hop, windowing_function=sp.hanning, dft_length=None,
         zp_flag=False):
    """ Calculates the Short-Time Fourier Transform a signal.

    Given an input signal, it calculates the DFT of frames of the signal and stores them
    in bi-dimensional Scipy array.

    **Args**:
        window_len (float): length of the window in seconds (must be positive).
        window (callable): a callable object that receives the window length in samples
                           and returns a numpy array containing the windowing function
                           samples.
        hop (float): frame hop between adjacent frames in seconds.
        final_time (positive integer): time (in seconds) up to which the spectrogram is
                                       calculated.
        zp_flag (bool): a flag indicating if the *Zero-Phase Windowing* should be
                        performed.

    **Returns**:
	spec: numpy array
	time: numpy array
	frequency: numpy array

    **Raises**:

    """
    # Checking input:
    if x.ndim != 1:
        raise AttributeError("Data must be one-dimensional.")
    # Window length must be odd:
    if window_length%2 == 0:
        window_length = window_length + 1
    # DFT length is equal the window_len+1 (always even)
    if dft_length is None:
        dft_length = window_length + 1
    # If dft_length was set by the user, it should always be larger than the window length.
    if dft_length < window_length:
        warnings.warn("DFT length is smaller than window length.", RuntimeWarning)
    # Partitioning the input signal:
    part_sig = segmentSignal(x, window_length, hop)
    no_cols = part_sig.shape[1]
    # Applying the window:
    window = windowing_function(window_length)
    win_sig = part_sig * sp.transpose(sp.tile(window, (no_cols, 1)))
    # Zero-phase windowing:
    if zp_flag:
        win_sig = fft.fftshift(win_sig, axes=0)
    # Taking the FFT of the partitioned signal
    spec = fft.fftshift(fft.fft(win_sig, n=dft_length, axis=0), axes=0)
    # Normalizing energy
    spec /= sp.sum(window)
    # Calculating time and frequency indices for the data
    frequency = fft.fftshift(fft.fftfreq(dft_length))
    time = sp.arange(no_cols)*float(hop) + ((window_length-1)/2)
    # Creating output spectrogram
    return spec, time, frequency


def segmentSignal(signal, window_len, hop):
    """ Segmentation of an array-like input:

    Given an array-like, this function calculates the DFT of frames of the signal and stores them
    in bi-dimensional Scipy array.


    **Args**:
        signal (array-like): object to be windowed. Must be a one-dimensional array-like object.
        window_len (int): window size in samples.
        hop (int): frame hop between adjacent frames in seconds.

    **Returns**:
        A 2-D numpy array containing the windowed signal. Each element of this array X
        can be defined as:

        X[m,n] = x[n+Hm]

        where, H is the HOP in samples, 0<=n<=N, N = window_len, and 0<m<floor(((len(x)-N)/H)+1).

    **Raises**:
        AttributeError if signal is not one-dimensional.
        ValueError if window_len or hop  are not strictly positives.
    """
    if(window_len <= 0 or hop <= 0):
        raise ValueError("window_len and hop values must be strictly positive numbers.")
    if signal.ndim != 1:
        raise AttributeError("Input signal must be one dimensional.")
    # Calculating the number of columns:
    no_cols = int(sp.floor((sp.size(signal)-window_len)/float(hop))+1)
    # Windowing indices (which element goes to which position in the windowed matrix).
    ind_col = sp.tile(sp.arange(window_len, dtype=np.uint64), (no_cols, 1))
    ind_line = sp.tile(sp.arange(no_cols, dtype=np.uint64)*hop, (window_len, 1))
    ind = sp.transpose(ind_col) + ind_line
    # Partitioned signal:
    part_sig = signal[ind].copy()
    # Windowing partitioned signal
    return part_sig


def fft2mel(freq, nfilts, minfreq, maxfreq):
    """ This method returns a 2-D Numpy array of weights that map a linearly spaced spectrogram
    to the Mel scale.

    **Args**:
        freq (1-D Numpy array): frequency of the components of the DFT.
        nfilts (): number of output bands.
        minfreq (): frequency of the first MEL coefficient.
        maxfreq (): frequency of the last MEL coefficient.

    **Returns**:
        The center frequencies in Hz of the Mel bands.

        """
    minmel = hz2mel(minfreq)
    maxmel = hz2mel(maxfreq)
    binfrqs = mel2hz(minmel+sp.arange(nfilts+2)/(float(nfilts)+1)*(maxmel-minmel))
    wts = sp.zeros((nfilts, (freq.size)))
    for i in range(nfilts):
        slp = binfrqs[i + sp.arange(3)]
        loslope = (freq - slp[0])/(slp[1] - slp[0])
        hislope = (slp[2] - freq)/(slp[2] - slp[1])
        wts[i, :] = sp.maximum(0.0, sp.minimum(loslope, hislope))
    wts[:, freq < 0] = 0
    wts = sp.dot(sp.diag(2./(binfrqs[2+sp.arange(nfilts)]-binfrqs[sp.arange(nfilts)])), wts)
    binfrqs = binfrqs[1:nfilts+1]

    return wts, binfrqs


def hz2mel(f_hz):
    """ Converts a given frequency in Hz to the Mel scale.

    **Args**:
        f_hz (Numpy array): Array containing the frequencies in HZ that should be converted.

    **Returns**:
        A Numpy array (of same shape as f_zh) containing the converted frequencies.

    """
    f_0 = 0
    f_sp = 200.0/3.0 # Log step
    brkfrq = 1000.0 # Frequency above which the distribution stays linear.
    brkpt = (brkfrq - f_0)/f_sp # First Mel value for linear region.
    logstep = sp.exp(sp.log(6.4)/27) # Step in the log region
    z_mel = sp.where(f_hz < brkfrq, (f_hz - f_0)/f_sp, brkpt +
                     (sp.log(f_hz/brkfrq))/sp.log(logstep))
    return z_mel


def mel2hz(z_mel):
    """ Converts a given frequency in the Mel scale to Hz scale.

    **Args**:
        z_mel (Numpy array): Array of frequencies in the Mel scale that should be converted.

    **Returns**:
        A Numpy array (of same shape as z_mel) containing the converted frequencies.
    """
    f_0 = 0
    f_sp = 200.0/3.0
    brkfrq = 1000.0
    brkpt = (brkfrq - f_0)/f_sp
    logstep = sp.exp(sp.log(6.4)/27)
    f_hz = sp.where(z_mel < brkpt, f_0 + f_sp*z_mel, brkfrq*sp.exp(sp.log(logstep)*(z_mel-brkpt)))
    return f_hz

def deltas(x, w=3):
    """ this function estimates the derivative of x
    """
    if x.ndim == 1:
        y = x.reshape((-1, 1)).T
    else:
        y = x
    if not w%2:
        w -= 1
    hlen = int(sp.floor(w/2))
    if w == 1: # first-order difference:
        win = sp.r_[1, -1]
    else:
        win = sp.r_[hlen:-hlen-1:-1]
    # Extending the input data (avoid border problems)
    extended_x = sp.c_[sp.repeat(y[:, 0].reshape((-1, 1)), hlen, axis=1), y,
                       sp.repeat(y[:, -1].reshape((-1, 1)), hlen, axis=1)]
    d = scipy.signal.lfilter(win, 1, extended_x, axis=1)
    d = d[:, 2*hlen:]
    if x.ndim == 1:
        d = d[0, :]
    return d


def getValidKeywords(kw, func):
    """ This function returns a dictionary containing the keywords arguments in initial_kw
        that are valid for function func.
    """
    import inspect
    valid_kw = {}
    invalid_kw = kw.copy()
    # args, varargs, varkw, defaults = inspect.getargspec(func)
    # NOTE: inspect.getargspec was deprecated, and inspect.getfullargspec is suggested as the
    # standard interface for single-source Python 2/3 code migrating away from legacy getargspec()
    args = inspect.getfullargspec(func)
    #for k, v in kw.iteritems():
    for k, v in kw.items():
        if k in args.args:
            valid_kw[k] = v
            del invalid_kw[k]
    return valid_kw, invalid_kw



# encoding: utf-8
# pylint: disable=C0103
# pylint: disable=too-many-arguments
"""Utility functions to deal with audio."""

import warnings
import numpy as np
import librosa.display
from matplotlib.cm import gray_r
from matplotlib.axes import Axes
from pylab import get_cmap
from matplotlib import colors
from matplotlib.ticker import Formatter
from . import util

__all__ = ['waveplot', 'specshow', 'mapshow']

# simply use librosa.specshow (this may change in the future)
specshow = librosa.display.specshow


def waveplot(y, sr=22050, x_axis='time', beats=None, beat_labs=None,
             ax=None, **kwargs):
    '''Plot an audio waveform and beat labels (optinal).


    Parameters
    ----------
    y : np.ndarray
        audio time series

    sr : number > 0 [scalar]
        sampling rate of `y`

    x_axis : str {'time', 'off', 'none'} or None
        If 'time', the x-axis is given time tick-marks.

    ax : matplotlib.axes.Axes or None
        Axes to plot on instead of the default `plt.gca()`.

    kwargs
        Additional keyword arguments to `matplotlib.`

    Returns
    -------

    See also
    --------


    Examples
    --------
    '''

    kwargs.setdefault('color', 'royalblue')
    kwargs.setdefault('linestyle', '-')
    kwargs.setdefault('alpha', 0.6)

    if y.ndim > 1:
        raise ValueError("`y` must be a one dimensional array. "
                         "Found y.ndim={}".format(y.ndim))

    # time array in seconds
    time = np.arange(y.size)/sr
    # its maximum value
    max_time = np.max(time)

    # check axes and create it if needed
    axes = __check_axes(ax)

    # plot waveform
    out = axes.plot(time, y, **kwargs)

    if beats is not None:
        __plot_beats(beats, max_time, axes, beat_labs=beat_labs, **kwargs)

    # format x axis
    if x_axis == 'time':
        axes.xaxis.set_major_formatter(TimeFormatter(lag=False))
        axes.xaxis.set_label_text('Time (s)')
    elif x_axis is None or x_axis in ['off', 'none']:
        axes.set_xticks([])
    else:
        raise ParameterError('Unknown x_axis value: {}'.format(x_axis))

    return out


def featureplot(feature, time, x_axis='time', beats=None, beat_labs=None,
                ax=None, **kwargs):
    '''Plot an audio waveform and beat labels (optinal).


    Parameters
    ----------
    feature : np.ndarray
        feature time series

    time : np.ndarray
        time instant of the feature values

    x_axis : str {'time', 'off', 'none'} or None
        If 'time', the x-axis is given time tick-marks.

    ax : matplotlib.axes.Axes or None
        Axes to plot on instead of the default `plt.gca()`.

    kwargs
        Additional keyword arguments to `matplotlib.`

    Returns
    -------

    See also
    --------


    Examples
    --------
    '''

    kwargs.setdefault('color', 'seagreen')
    kwargs.setdefault('linestyle', '-')
    kwargs.setdefault('alpha', 0.8)

    if feature.ndim > 1:
        raise ValueError("`feature` must be a one dimensional array. "
                         "Found feature.ndim={}".format(feature.ndim))

    # maximum time value
    max_time = np.max(time)

    # check axes and create it if needed
    axes = __check_axes(ax)

    # plot waveform
    out = axes.plot(time, feature, **kwargs)

    if beats is not None:
        __plot_beats(beats, max_time, axes, beat_labs=beat_labs, **kwargs)

    # format x axis
    if x_axis == 'time':
        axes.xaxis.set_major_formatter(TimeFormatter(lag=False))
        axes.xaxis.set_label_text('Time (s)')
    elif x_axis is None or x_axis in ['off', 'none']:
        axes.set_xticks([])
    else:
        raise ParameterError('Unknown x_axis value: {}'.format(x_axis))

    return out


def __plot_beats(beats, max_time, ax, beat_labs=None, **kwargs):
    '''Plot beat labels.


    Parameters
    ----------
    beats : np.ndarray
        audio time series

    beat_labs : list
        beat labels

    x_axis : str {'time', 'off', 'none'} or None
        If 'time', the x-axis is given time tick-marks.

    ax : matplotlib.axes.Axes or None
        Axes to plot on instead of the default `plt.gca()`.

    kwargs
        Additional keyword arguments to `matplotlib.`

    Returns
    -------

    See also
    --------


    Examples
    --------
    '''

    kwargs['color'] = 'black'
    kwargs.setdefault('linestyle', '-')
    kwargs['alpha'] = 0.3
    kwargs.setdefault('linewidth', 2)

    # consider beats (and labels) bellow max_time
    ind_beat = util.find_nearest(beats, max_time)
    new_beats = beats[:ind_beat]
    if beat_labs is not None:
        new_labs = beat_labs[:ind_beat]

    # plot beat annotations
    for beat in new_beats:
        ax.axvline(x=beat, **kwargs)

    # set ticks and labels
    ax2 = ax.twiny()
    ax2.set_xlim(ax.get_xlim())
    ax2.set_xticks(new_beats)
    ax2.set_xticklabels(new_labs)
    #ax2.set_xlabel("beats")

    return ax2


def mapshow(data, x_coords=None, y_coords=None, ax=None,
            n_tatums=4, clusters=None, **kwargs):
    '''Display a feature map.

    Parameters
    ----------
    data : np.ndarray
        Feature map to display

    x_coords : np.ndarray [shape=data.shape[1]+1]
    y_coords : np.ndarray [shape=data.shape[0]+1]

        Optional positioning coordinates of the input data.

    ax : matplotlib.axes.Axes or None
        Axes to plot on instead of the default `plt.gca()`.

    n_tatums : int
        Number of tatums (subdivisions) per tactus beat

    clusters : np.ndarray
        Array indicating cluster number for each pattern of the input data.
        If provided (not None) the clusters area displayed with colors.

    kwargs : additional keyword arguments
        Arguments passed through to `matplotlib.pyplot.pcolormesh`.

        By default, the following options are set:

            - `cmap=gray_r`
            - `rasterized=True`
            - `edgecolors='None'`
            - `shading='flat'`

    Returns
    -------
    axes
        The axis handle for the figure.


    See Also
    --------
    matplotlib.pyplot.pcolormesh


    Examples
    --------

    '''

    kwargs.setdefault('cmap', gray_r)
    kwargs.setdefault('rasterized', True)
    kwargs.setdefault('edgecolors', 'None')
    kwargs.setdefault('shading', 'flat')

    # number of bars
    bars = data.shape[0]
    # number of tatums in a bar
    tatums = data.shape[1]

    # set the x and y coordinates
    y_coords = np.array(range(tatums+1))+0.5
    x_coords = np.array(range(bars))+1

    # check axes and create it if needed
    axes = __check_axes(ax)
    # plot rhythmic patterns map (grayscale)
    out = axes.pcolormesh(x_coords, y_coords, data.T, **kwargs)
    __set_current_image(ax, out)

    # if clusters are given then show them in colors
    if clusters is not None:
        # check clusters and return number of clusters
        n_clusters = __check_clusters(clusters, bars)
        # matrix and other elements needed to plot clusters' map
        mapc, cmap, norm = __get_cluster_matrix(clusters, n_clusters, y_coords.size)
        # plot clusters in colors
        axes.pcolormesh(x_coords, y_coords, mapc, cmap=cmap, norm=norm, alpha=0.6)

    # set axes limits
    axes.set_xlim(x_coords.min()-0.5, x_coords.max()+0.5)
    axes.set_ylim(y_coords.min(), y_coords.max())


    # configure tickers and labels
    __decorate_axis_map(axes, tatums=n_tatums)

    return axes


def __check_axes(axes):
    '''Check if "axes" is an instance of an axis object.'''
    if axes is None:
        import matplotlib.pyplot as plt
        axes = plt.gca()
    elif not isinstance(axes, Axes):
        raise ValueError("`axes` must be an instance of matplotlib.axes.Axes. "
                         "Found type(axes)={}".format(type(axes)))
    return axes

def __check_clusters(clusters, bars):
    '''Check if "clusters" is an instance of an axis object.
       Check if "clusters" is a one dimensional array of the correct length.
       '''
    if isinstance(clusters, np.ndarray):
        if clusters.ndim == 1:
            if clusters.size == bars:
                # count number of clusters
                n_clusters = np.unique(clusters).size
        else:
            raise ValueError("`clusters` must be a one dimensional array. "
                             "Found clusters.ndim={}".format(clusters.ndim))
    else:
        raise ValueError("`clusters` must be an instance of numpy.ndarray. "
                         "Found type(axes)={}".format(type(clusters)))

    return n_clusters


def __get_cluster_matrix(clusters, n_clusters, n_tatums):
    '''Get clusters' matrix and other elements needed to plot clusters' map.'''
    # make a color map of fixed colors for colormesh
    # cmap = get_cmap('RdBu', n_clusters)
    cmap = get_cmap('tab10', n_clusters)
    bounds = range(n_clusters+1)
    norm = colors.BoundaryNorm(bounds, cmap.N)
    mapc = np.tile(clusters+0.5, (n_tatums, 1))

    return mapc, cmap, norm


def __set_current_image(ax, img):
    '''Helper to set the current image in pyplot mode.

    If the provided `ax` is not `None`, then we assume that the user is using the object API.
    In this case, the pyplot current image is not set.
    '''

    if ax is None:
        import matplotlib.pyplot as plt
        plt.sci(img)


def __decorate_axis_map(axis, tatums=4):
    '''Configure axis ticks and labels for feature map plot'''

    # ticks at beats
    ylims = axis.get_ylim()
    all_tatums = int(ylims[1])
    ticks_beats = [x+0.5 for x in range(0, all_tatums, tatums)]
    num_beats = int(all_tatums / tatums)
    labels_beats = [x+1 for x in range(num_beats)]

    axis.yaxis.set_ticks(ticks_beats)
    axis.set_yticklabels(labels_beats)
    axis.tick_params(labelsize=10)
    # axis.yaxis.set_major_formatter(NullFormatter())
    axis.yaxis.grid()
    gridlines = axis.get_ygridlines()
    for line in gridlines:
        line.set_linestyle('-')
        line.set_linewidth(2)
        line.set_color('black')
    axis.set_ylabel('beats')


class TimeFormatter(Formatter):
    '''A tick formatter for time axes. Copied from librosa.

    Automatically switches between seconds, minutes:seconds,
    or hours:minutes:seconds.

    Parameters
    ----------
    lag : bool
        If `True`, then the time axis is interpreted in lag coordinates.
        Anything past the midpoint will be converted to negative time.

    unit : str or None
        Abbreviation of the physical unit for axis labels and ticks.
        Either equal to `s` (seconds) or `ms` (milliseconds) or None (default).
        If set to None, the resulting TimeFormatter object adapts its string
        representation to the duration of the underlying time range:
        `hh:mm:ss` above 3600 seconds; `mm:ss` between 60 and 3600 seconds;
        and `ss` below 60 seconds.


    See also
    --------
    matplotlib.ticker.Formatter


    Examples
    --------

    For normal time

    >>> import matplotlib.pyplot as plt
    >>> times = np.arange(30)
    >>> values = np.random.randn(len(times))
    >>> plt.figure()
    >>> ax = plt.gca()
    >>> ax.plot(times, values)
    >>> ax.xaxis.set_major_formatter(librosa.display.TimeFormatter())
    >>> ax.set_xlabel('Time')

    Manually set the physical time unit of the x-axis to milliseconds

    >>> times = np.arange(100)
    >>> values = np.random.randn(len(times))
    >>> plt.figure()
    >>> ax = plt.gca()
    >>> ax.plot(times, values)
    >>> ax.xaxis.set_major_formatter(librosa.display.TimeFormatter(unit='ms'))
    >>> ax.set_xlabel('Time (ms)')

    For lag plots

    >>> times = np.arange(60)
    >>> values = np.random.randn(len(times))
    >>> plt.figure()
    >>> ax = plt.gca()
    >>> ax.plot(times, values)
    >>> ax.xaxis.set_major_formatter(librosa.display.TimeFormatter(lag=True))
    >>> ax.set_xlabel('Lag')
    '''

    def __init__(self, lag=False, unit=None):

        if unit not in ['s', 'ms', None]:
            raise ParameterError('Unknown time unit: {}'.format(unit))

        self.unit = unit
        self.lag = lag

    def __call__(self, x, pos=None):
        '''Return the time format as pos'''

        _, dmax = self.axis.get_data_interval()
        vmin, vmax = self.axis.get_view_interval()

        # In lag-time axes, anything greater than dmax / 2 is negative time
        if self.lag and x >= dmax * 0.5:
            # In lag mode, don't tick past the limits of the data
            if x > dmax:
                return ''
            value = np.abs(x - dmax)
            # Do we need to tweak vmin/vmax here?
            sign = '-'
        else:
            value = x
            sign = ''

        if self.unit == 's':
            s = '{:.3g}'.format(value)
        elif self.unit == 'ms':
            s = '{:.3g}'.format(value * 1000)
        else:
            if vmax - vmin > 3600:
                s = '{:d}:{:02d}:{:02d}'.format(int(value / 3600.0),
                                                int(np.mod(value / 60.0, 60)),
                                                int(np.mod(value, 60)))
            elif vmax - vmin > 60:
                s = '{:d}:{:02d}'.format(int(value / 60.0),
                                         int(np.mod(value, 60)))
            else:
                s = '{:.2g}'.format(value)

        return '{:s}{:s}'.format(sign, s)
